/*
 * uart1.c : serial1 handling routines with interrupts
 *   by uratan! 2008.9.5
 */
#include "common.h"
#include "LPC214x.h"
#include "irq.h"

/*
 * switch uart0/uart1 for debug
 *   (you shall care about debug-uart on uart0)
 */
#define USE_UART1   1

/*
 * my local debug flag
 */
#define myUDEBUG  (0 && gUDEBUG)
#define PUTCHAR    if(myUDEBUG && USE_UART1)cm_putchar

/*
 * choise FIFO depth
 */
#define N_BUFF          16		/* must be 2^x */
#define BUFF_PTR_MASK  (N_BUFF-1)

/*
 * switch registers
 */
#if USE_UART1
# define UxDLL U1DLL
# define UxDLM U1DLM
# define UxFCR U1FCR
# define UxFDR U1FDR
# define UxIER U1IER
# define UxIIR U1IIR
# define UxLCR U1LCR
# define UxLCR U1LCR
# define UxLSR U1LSR
# define UxRBR U1RBR
# define UxTHR U1THR
# define UARTx_INT  UART1_INT
#else /* USE_UART1 */
# define UxDLL U0DLL
# define UxDLM U0DLM
# define UxFCR U0FCR
# define UxFDR U0FDR
# define UxIER U0IER
# define UxIIR U0IIR
# define UxLCR U0LCR
# define UxLCR U0LCR
# define UxLSR U0LSR
# define UxRBR U0RBR
# define UxTHR U0THR
# define UARTx_INT  UART0_INT
#endif /* USE_UART1 */

/*
 * register bitmaps
 */
/* UxIER: Interrupt Enable Register */
#define IER_RBR		0x01	/* receiver buffer register(RDA) + CTI */
#define IER_THRE	0x02	/* transmit holding register */
#define IER_RLS		0x04	/* RX line status */
/* UxIIR: Interrupt Identification Register */
#define IIR_PEND	0x01
#define IIR_ID_MASK	(0x07<<1)	/* mask for intr ID */
#define   ID_RLS	(0x03<<1)	/* Receive Line Status */
#define   ID_RDA	(0x02<<1)	/* Receive Data Available */
#define   ID_CTI	(0x06<<1)	/* Character Time-out Indicator */
#define   ID_THRE	(0x01<<1)	/* THRE interrupt */
/* UxLSR: Line Status Register */
#define LSR_RDR		0x01	/* Receiver Data Ready */
#define LSR_OE		0x02	/* Overrun Error */
#define LSR_PE		0x04	/* Parity Error */
#define LSR_FE		0x08	/* Framing Error */
#define LSR_BI		0x10	/* Break Interrupt */
#define LSR_THRE	0x20	/* Transmitter Holding Register Empty */
#define LSR_TEMT	0x40	/* Transmitter Empty */
#define LSR_RXFE	0x80	/* Error in RX FIFO */

/*
 * functions
 */
int uart1_init(u_long baudrate);
int uart1_peek_recv(void);
void uart1_set_to_send(u_char c);
static void uart1_handler(void) __attribute__ ((interrupt("IRQ")));

/*
 * valiables
 */
static u_char send_buff[N_BUFF];
static u_long send_in_ptr, send_out_ptr;
static VOLATILE u_long n_send;		/* need 'VOLATILE' because pollig this */
static int ti1_will_come;

static u_char recv_buff[N_BUFF];
static u_long recv_in_ptr, recv_out_ptr;
static u_long n_recv;


/*
 * initial serial1
 */
int uart1_init(u_long baudrate)
{
	u_long Fdiv;

	PCONP |= PP_PCUART1;			/* turn uart1 on */

#if USE_UART1
	PINSEL0 |= (1<<18) | (1<<16);	/* enable RxD1 and TxD1 */
#else /* USE_UART1 */
	PINSEL0 |= (1<<2) | (1<<0);		/* enable RxD0 and TxD0 */
#endif /* USE_UART1 */

	UxIER = 0;				/* once kill interrupts */

	UxLCR = 0x80;			/* DLAB=1 */
	UxFDR = 0x10;			/* do not use Fractional Divider */
	Fdiv = (Fpclk / 16) / baudrate;		/* calc division from baudrate */
	UxDLM = Fdiv / 256;
	UxDLL = Fdiv % 256;
	UxLCR = 0x03;			/* DLAB=0, 8bits, no-parity, 1stopbit */

//	UxFCR = 0x07;			/* Enable and reset TX and RX FIFO. */
	UxFCR = 0x00;			/* do not use FIFO (inhibited by datasheet!) */

	/*
	 * initial both FIFO
	 */
	send_in_ptr = send_out_ptr = 0;
	n_send = 0;
	ti1_will_come = FALSE;		/* no THRE intr will come */

	recv_in_ptr = recv_out_ptr = 0;
	n_recv = 0;

	/*
	 * setup vectored interrupt
	 */
	if(install_irq(UARTx_INT, (void *)uart1_handler) == FALSE) {
		return(FALSE);
	}

	/*
	 * at last, enable interrupts
	 */
	UxIER = IER_RBR | IER_THRE;		/* enable interrupts */

	return(TRUE);
}

/*
 * check receive buffer
 */
int uart1_peek_recv(void)
{
	u_long imf;
	u_char c;

	/*
	 * chech buffer empty
	 */
	if(n_recv == 0)
		return(-1);		/* no data has come */

	/*
	 * take 1char out from buffer
	 */
	imf = IntDisable();				/* disable interrupts */

	c = recv_buff[recv_out_ptr++];
	n_recv--;
	recv_out_ptr &= BUFF_PTR_MASK;

	IntRestore(imf);				/* (may) enable interrupts */

	return(c);
}

/*
 * pile the data on send buffer
 */
void uart1_set_to_send(u_char c)
{
	u_long imf;

	/*
	 * wait if there is no space in send buffer,
	 * i.e, keep host waiting on process of SET_REPORT request
	 */
	while(n_send == N_BUFF)
		;

	/*
	 * try to set the char
	 */
	imf = IntDisable();				/* disable interrupts */

	if(ti1_will_come) {
		/*
		 * I am doing transfer now, so THRE intr will flush the buffer
		 */
		send_buff[send_in_ptr++] = c;
		n_send++;
		send_in_ptr &= BUFF_PTR_MASK;
PUTCHAR('s');
	} else {
		/*
		 * transmitter is idle, no THRE intr will come, so
		 */
		UxTHR = c;			/* send directly here */
		ti1_will_come = TRUE;
PUTCHAR('d');
	}

	IntRestore(imf);				/* (may) enable interrupts */

	return;
}

/*
 * uart1 interrupt service routine
 */
static void uart1_handler(void)
{
	u_char iid, c;

	/*
	 * take out interrupt reason
	 */
	iid = UxIIR & IIR_ID_MASK;		/* Interrupt Identification */

	if((iid == ID_RDA) || (iid == ID_CTI)) {
		/*
		 * in case of Receive Data Available, or Character Time-out indicator
		 */
PUTCHAR('R');
		/*
		 * read the data, then put it on receive buffer
		 */
		c = UxRBR;			/* <== this clears interrupt request */

		if(n_recv == N_BUFF) {
			/*
			 * buffer overflow, only ignore
			 */
PUTCHAR('O');
		} else {
			/*
			 * save received data on buffer
			 */
			recv_buff[recv_in_ptr++] = c;
			n_recv++;
			recv_in_ptr &= BUFF_PTR_MASK;
		}

		VICVectAddr = 0;		/* Acknowledge Interrupt */
		return;
	}

	if(iid == ID_THRE) {
		/*
		 * in case of THRE, transmit holding register empty
		 *   the request is automatically cleared when reading UxIIR above
		 */
PUTCHAR('T');
		/*
		 * confirm status in the LSR
		 */
		if(UxLSR & LSR_THRE) {
			if(n_send == 0) {
				/*
				 * flushing buffer is done, THRE intr chain is stopped here
				 */
				ti1_will_come = FALSE;
			} else {
				/*
				 * send data on buffer
				 *   this will generate next THRE intr
				 */
				UxTHR = send_buff[send_out_ptr++];
				n_send--;
				send_out_ptr &= BUFF_PTR_MASK;
			}
		}

		VICVectAddr = 0;		/* Acknowledge Interrupt */
		return;
	}

	/*
	 * something is wrong !!
	 */
PUTCHAR('X');
	while(1)
		;

	VICVectAddr = 0;		/* Acknowledge Interrupt */
}
